;***********************************************************************************;
;***********************************************************************************;
;
; The almost completely commented VIC-1112 ROM disassembly. v1.0
; By Simon Rowe <srowe@mose.org.uk>.

; This is a bit correct assembly listing for the ROM in the VIC-20 IEEE-488 cartridge.


;***********************************************************************************;
;***********************************************************************************;
;
; KERNAL zero page

status	= $90
verck	= $93
bsour	= $95
ldtnd	= $98
dfltn	= $99
dflto	= $9a
sal	= $ac
eal	= $ae
fnlen	= $b7
la	= $b8
sa	= $b9
fa	= $ba
fnadr	= $bb
memuss	= $c3

lat	= $0259
fat	= $0263
sat	= $026d

	; KERNAL vectors
iopen	= $031a
iload	= $0330
isave	= $0332
igetin	= $032a
nminv	= $0318

	; Standard VIA
via1pa1	= $9111			; VIA 1 DRA
					; bit	function
					; ---	--------
					;  7	serial ATN out
					;  6	cassette switch
					;  5	joystick fire, light pen
					;  4	joystick left, paddle X fire
					;  3	joystick down
					;  2	joystick up
					;  1	serial DATA in
					;  0	serial CLK in

via1ifr	= $911d			; VIA 1 IFR
					; bit	function		cleared by
					; ---	--------		----------
					;  7	interrupt status	clearing all enabled interrupts
					;  6	T1 interrupt		read T1C_l, write T1C_h
					;  5	T2 interrupt		read T2C_l, write T2C_h
					;  4	CB1 transition		read or write port B
					;  3	CB2 transition		read or write port B
					;  2	8 shifts done		read or write the shift register
					;  1	CA1 transition		read or write port A
					;  0	CA2 transition		read or write port A

via1ier	= $911e			; VIA 1 IER
					; bit	function
					; ---	--------
					;  7	enable/disable
					;  6	T1 interrupt
					;  5	T2 interrupt
					;  4	CB1 transition
					;  3	CB2 transition
					;  2	8 shifts done
					;  1	CA1 transition
					;  0	CA2 transition

	; VIA A (U4)
viaapb	= $9800			; VIA A DRB
					; bit	function
					; ---	--------
					;  7	ATN in
					;  6	NDAC in
					;  5	NRFD in
					;  4	DAV in
					;  3	EOI in
					;  2	NDAC out
					;  1	NRFD out
					;  0	DAV out

viaaddrb = $9802		; VIA A DDRB
viaaddra = $9803		; VIA A DDRA
viaaacr	= $980b			; VIA A ACR
					; bit	function
					; ---	--------

viaapcr	= $980c			; VIA A PCR
					; bit	function
					; ---	--------

viaaifr = $980d			; VIA A IFR
viaaier	= $980e			; VIA A IER
viaapa2	= $980f			; VIA A DRA, no handshake

	; VIA B (U5)
viabpb	= $9810			; VIA B DRB
					; bit	function
					; ---	--------
					;  7	DIO8 in
					;  6	DIO7 in
					;  5	DIO6 in
					;  4	DIO5 in
					;  3	DIO4 in
					;  2	DIO3 in
					;  1	DIO2 in
					;  0	DIO1 in

viabddrb = $9812		; VIA B DDRB
viabddra = $9813		; VIA B DDRA
viabt1ch = $9815		; VIA B T1C_h
viabacr	= $981b			; VIA B ACR
viabpcr	= $981c			; VIA B PCR
					; bit	function
					; ---	--------
					; 765	CB2 EOI out
					;  4	CB1 SRQ in
					; 321	CA2 ATN out
					;  0	

viabifr	= $981d			; VIA B IFR
viabier	= $981e			; VIA B IER
viabpa2	= $981f			; VIA B DRA, no handshake
					; bit	function
					; ---	--------
					;  7	DIO8 out
					;  6	DIO7 out
					;  5	DIO6 out
					;  4	DIO5 out
					;  3	DIO4 out
					;  2	DIO3 out
					;  1	DIO2 out
					;  0	DIO1 out

	; Unknown registers on IO3/ *** dead code ***
unkwn1	= $9c00
unkwn2	= $9c01

	; XROM entrypoints
xromwarm = $a002
	
	; BASIC entrypoints
warmst	= $c002
ready	= $c474
prtstr	= $cb1e

	; KERNAL entrypoints
initsk	= $e518
scrnout	= $e742
rsdvcerr = $f0b9
chrinsr	= $f264
chrinrs	= $f26f
fndflno	= $f3cf
setflch	= $f3df
srching	= $f647
ldvrmsg	= $f66a
saving	= $f728
fudtim	= $f734
fe_2mnyf = $f77e
fe_alopn = $f781
fe_ntopn = $f784
fe_ntfnd = $f787
fe_dvntp = $f78a
fe_ntinp = $f78d
fe_ntout = $f790
fe_misfn = $f793
fe_ildev = $f796
rd300	= $fbd2
vprty	= $fd11
wrt62	= $fd1b
chkauto	= $fd3f
frestor	= $fd52
initvia	= $fdf9
oriost	= $fe6a
nmi2	= $fead
rsnmi	= $fede
_rti	= $ff56
stop	= $ffe1
	
	* = $b000

	.word einit
	.word enminv
	.byte "B0"
	nop
	nop
	nop

einit	jsr intevia		; initialize cartridge I/O registers
	jsr initvec		; initialize KERNAL vectors to IEEE-488 functions
	lda #<banner		; set banner pointer low byte
	ldy #>banner		; set banner pointer high byte
	jsr prtstr		; print banner
	ldx #$fb		; value for start stack
	txs			; set stack pointer
	jmp ready		; return to BASIC

banner	.byte $0d,"-- IEEE-488 DRIVER --",$0d,$00


;***********************************************************************************;
;
; initialize KERNAL vectors

initvec	ldx #$0d		; set index to vector table length
lb036	lda vtable,x		; get vector
	sta iopen,x		; save vector
	dex			; decrement index 
	bpl lb036		; loop until index < 0
	lda #<eload		; get ELOAD address low byte
	sta iload		; save to ILOAD vector low byte
	lda #>eload		; get ELOAD address high byte
	sta iload+1		; save to ILOAD vector high byte
	lda #<esave		; get ESAVE address low byte
	sta isave		; save to ISAVE vector low byte
	lda #>esave		; get ESAVE address high byte
	sta isave+1		; save to ISAVE vector high byte
	lda #<egetin		; get EGETIN address low byte
	sta igetin		; save to IGETIN vector low byte
	lda #>egetin		; get EGETIN address high byte
	sta igetin+1		; save to IGETIN vector high byte
	lda #<enminv		; get ENMINV address low byte
	sta nminv		; save to NMINV vector low byte
	lda #>enminv		; get ENMINV address high byte
	sta nminv+1		; save to NMINV vector high byte
	rts

vtable	.word eopen		; OPEN vector
	.word eclose		; CLOSE vector
	.word echkin		; CHKIN vector 
	.word echkout		; CHKOUT vector
	.word eclrchn		; CLRCHN vector
	.word echrin		; CHRIN vector
	.word echrout		; CHROUT vector


;***********************************************************************************;
;
; NMI handler

enminv	sei			; disable interrupts
	pha			; save .A
	txa			; copy .X
	pha			; save .X
	tya			; copy .Y
	pha			; save .Y
	lda via1ifr		; get VIA 1 IFR
	bmi lb084		; branch if interrupt raised
	jmp _rti		; restore registers and exit interrupt
lb084	and via1ier		; AND with VIA 1 IER
	tax			; copy to .X
	and #$02		; mask [RESTORE] key
	bne lb08f		; branch if [RESTORE] ket
	jmp rsnmi		; jump to KERNAL to handle RS-232 interrupt
lb08f	jsr chkauto		; scan for autostart ROM at $A000
	bne lb097		; branch if no autostart ROM
	jmp (xromwarm)		; else do autostart ROM break entry
lb097	bit via1pa1		; test VIA 1 DRA
	jsr fudtim		; increment the real time clock
	jsr stop		; scan stop key
	beq ebreak		; branch if not [STOP]
	jmp _rti		; restore registers and exit interrupt


;***********************************************************************************;
;
; BRK handler

ebreak	jsr frestor		; restore default I/O vectors
	jsr initvia		; initialize I/O registers
	jsr initsk		; initialize hardware
	jsr intevia		; initialize cartridge I/O registers
	jsr initvec		; initialize KERNAL vectors to IEEE-488 functions
	jmp (warmst)		; do BASIC break entry


;***********************************************************************************;
;
; open a channel for output

; Any logical file that has already been opened by the OPEN routine can be defined
; as an output channel by this routine the device on the channel must be an output
; output device or an error will occur and the routine will abort.

; If you are sending data to anywhere other than the screen this routine must be
; called before using the CHROUT routine. If you are sending data to the screen and
; no other output channels are open then the calls to this routine and to the OPEN
; routine are not needed.

; When used with a device on the IEEE bus this routine will automatically send the
; listen address specified by the OPEN routine and any secondary address.

; Possible errors are:
;
;	3 : file not open
;	5 : device not present
;	7 : file is not an output file

echkout	jsr fndflno		; find file
	beq lb0bf		; branch if file found
	jmp fe_ntopn		; do file not open error and retirn
lb0bf	jsr setflch		; set file details from table,X
	lda fa			; get device number
	bne lb0c9		; branch if device is not keyboard
	jmp fe_ntout		; do not output file error and return
lb0c9	cmp #$03		; compare device number with screen
	beq lb0d2		; if screen save output device number and exit
	bcs lb0d6		; branch if > screen, IEEE bus device
	jmp $f321		; jump to KERNAL to handle RS-232 and tape
lb0d2	sta dflto		; save output device number
	clc			; flag ok
	rts
lb0d6	jmp lb0e8		; skip dead code

	nop			; *** dead code ***
	jmp $f332		; *** dead code ***
	cmp #$09		; *** dead code ***
	bne lb0e8		; *** dead code ***
	jsr lb59d		; *** dead code ***
	sta dflto		; *** dead code ***
	clc			; *** dead code ***
	rts			; *** dead code ***

lb0e8	jsr lb40c		; command an IEEE device to LISTEN
	sta dflto		; save output device number
	clc			; flag ok
	rts


;***********************************************************************************;
;
; output a character to channel

; This routine will output a character to an already opened channel. Use the OPEN
; routine, OPEN, and the CHKOUT routine, to set up the output channel before calling
; this routine. If these calls are omitted, data will be sent to the default output
; device, device 3, the screen. The data byte to be output is loaded into the accumulator,
; and this routine is called. The data is then sent to the specified output device.
; The channel is left open after the call.

; NOTE: Care must be taken when using routine to send data to an IEEE device since
; data will be sent to all open output channels on the bus. Unless this is desired,
; all open output channels on the IEEE bus other than the actually intended
; destination channel must be closed by a call to the KERNAL close channel routine.

echrout	pha			; save the character to send
	lda dflto		; get output device number
	cmp #$03		; compare device number with screen
	bne lb0fa		; if output device not screen continue
	pla			; restore character to send
	jmp scrnout		; output character and return
lb0fa	bcs lb0ff		; branch if IEEE bus device
	jmp $f28b		; jump to KERNEL to handle RS-232 and tape
lb0ff	jmp lb112		; skip dead code

	nop			; *** dead code ***
	jmp $f287		; *** dead code ***
	cmp #$09		; *** dead code ***
	bne lb112		; *** dead code ***
	pla			; *** dead code ***
	sta bsour		; *** dead code ***
	jsr lb5ac		; *** dead code ***
	clc			; *** dead code ***
	rts			; *** dead code ***

lb112	pla			; restore character to send
	sta bsour		; store byte to send
	jsr lb428		; send byte on IEEE bus
	lda bsour		; restore byte sent
	clc			; flag ok
	rts


;***********************************************************************************;
;
; open a channel for input

; Any logical file that has already been opened by the OPEN routine can be defined as
; an input channel by this routine. The device on the channel must be an input device
; or an error will occur and the routine will abort.

; If you are getting data from anywhere other than the keyboard, this routine must be
; called before using either the CHRIN routine or the GETIN routine. If you are
; getting data from the keyboard and no other input channels are open then the calls
; to this routine and to the OPEN routine are not needed.

; When used with a device on the IEEE bus this routine will automatically send the
; listen address specified by the OPEN routine and any secondary address.

; Possible errors are:
;
;	3 : file not open
;	5 : device not present
;	6 : file is not an input file

echkin	jsr fndflno		; find file
	beq lb124		; branch if file opened
	jmp fe_ntopn		; do file not open error and return
lb124	jsr setflch		; set file details from table,X
	lda fa			; get device number
	beq lb134		; if device was keyboard save device #, flag ok and exit
	cmp #$03		; compare device number with screen
	beq lb134		; if device was screen save device #, flag ok and exit
	bcs lb138		; branch if IEEE bus device
	jmp $f2dc		; jump to KERNAL to handle RS-232 and tape
lb134	sta dfltn		; save input device number
	clc			; flag ok
	rts
lb138	jmp lb146		; skip dead code

	nop			; *** dead code ***
	jmp $f2f0		; *** dead code ***
	cmp #$09		; *** dead code ***
	bne lb146		; *** dead code ***
	jmp lb610		; *** dead code ***

lb146	jsr lb34d		; command an IEEE device to TALK
	sta dfltn		; save input device number
	clc			; flag ok
	rts


;***********************************************************************************;
;
; input character from channel

; This routine will get a byte of data from the channel already set up as the input
; channel by the CHKIN routine.

; If CHKIN has not been used to define another input channel the data is expected to be
; from the keyboard. the data byte is returned in the accumulator. The channel remains
; open after the call.

echrin	lda dfltn		; get input device number
	bne lb154		; if it's not the keyboard continue
	jmp $f212		; jump to KERNAL to handle keyboard input
lb154	cmp #$03		; compare device number with screen
	bne lb15b		; if it's not the screen continue
	jmp $f221		; jump to KERNAL to handle screen input
lb15b	bcs lb160		; branch if IEEE bus
	jmp $f22c		; jump to KERNAL to handle RS-232 and tape
lb160	jmp lb16e		; skip dead code

	nop			; *** dead code ***
	jmp chrinsr		; *** dead code ***
	cmp #$09		; *** dead code ***
	bne lb16e		; *** dead code ***
	jmp chrinrs		; *** dead code ***

lb16e	lda status		; get I/O status byte
	beq lb176		; if no errors flagged go input byte and return
	lda #$0d		; else return [EOL]
	clc			; flag ok
	rts
lb176	jsr lb44f		; input a byte from the IEEE bus
	lda bsour		; get received byte
	clc			; flag ok
	rts


;***********************************************************************************;
;
; close input and output channels

; This routine is called to clear all open channels and restore the I/O channels to
; their original default values. It is usually called after opening other I/O
; channels and using them for input/output operations. The default input device is
; 0, the keyboard. The default output device is 3, the screen.

; If one of the channels to be closed is to the IEEE bus, an UNTALK signal is sent
; first to clear the input channel or an UNLISTEN is sent to clear the output channel.
; By not calling this routine and leaving listener(s) active on the IEEE bus,
; several devices can receive the same data from the VIC at the same time. One way to
; take advantage of this would be to command the printer to LISTEN and the disk to
; TALK. This would allow direct printing of a disk file.

eclrchn	ldx #$03		; set .X to screen
	cpx dflto		; compare output device number with screen
	bcs lb198		; branch if >= screen
	nop			; else was IEEE bus
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	jsr lb3a1		; command the IEEE bus to UNLISTEN
	jmp lb198		; skip dead code

	jsr $ffae		; *** dead code ***

lb198	ldx #$03		; set .X to screen
	cpx dfltn		; compare input device number with screen
	bcs lb1b3		; branch if >= screen
	nop			; else was IEEE bus
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	jsr lb3aa		; send UNTALK command
	jmp lb1b3		; set default devices

	jsr $eef9		; *** dead code ***

lb1b3	ldx #$03		; set .X to screen
	stx dflto		; set output device number to screen
	lda #$00		; set .X to keyboard
	sta dfltn		; set input device number to keyboard
	rts


;***********************************************************************************;
;
; close a specified logical file

; This routine is used to close a logical file after all I/O operations have been
; completed on that file. This routine is called after the accumulator is loaded
; with the logical file number to be closed, the same number used when the file was
; opened using the OPEN routine.

eclose	jsr $f3d4		; find file .A
	beq lb1c3		; if the file is found go close it
	clc			; else the file was closed so just flag ok
	rts

;  found the file so close it

lb1c3	jsr setflch		; set file details from table,X
	txa			; copy file index to .A
	pha			; save file index
	lda fa			; get device number
	beq lb1d5		; if $00, keyboard, restore index and close file
	cmp #$03		; compare device number with screen
	beq lb1d5		; if screen restore index and close file
	bcs lb1d8		; if > screen go do IEEE bus device close
	jmp $f360		; else jump to KERNAL to handle RS-232 and tape
lb1d5	jmp $f3b1		; close file index .X
lb1d8	jmp lb1e6		; skip dead code

	nop			; *** dead code ***
	jmp $f3ae		; *** dead code ***
	cmp #$09		; *** dead code ***
	bne lb1e6		; *** dead code ***
	jmp lb5f6		; *** dead code ***

lb1e6	jsr lb3bb		; close IEEE bus device
	jmp $f3b1		; close file index .X


;***********************************************************************************;
;
; open a logical file

; This routine is used to open a logical file. Once the logical file is set up it
; can be used for input/output operations. Most of the I/O KERNAL routines call on
; this routine to create the logical files to operate on. No arguments need to be
; set up to use this routine, but both the SETLFS and SETNAM KERNAL routines must
; be called before using this routine.

eopen	ldx la			; get logical file
	bne lb1f3		; branch if there is a file
	jmp fe_ntinp		; else do not input file error and return
lb1f3	jsr fndflno		; find file
	bne lb1fb		; branch if file not found
	jmp fe_alopn		; else do file already open error and return
lb1fb	ldx ldtnd		; get open file count
	cpx #$0a		; compare with max
	bcc lb204		; branch if less
	jmp fe_2mnyf		; else do too many files error and return
lb204	inc ldtnd		; increment open file count
	lda la			; get logical file
	sta lat,x		; save to logical file table
	lda sa			; get secondary address
	ora #$60		; OR with the OPEN CHANNEL command
	sta sa			; set secondary address
	sta sat,x		; save to secondary address table
	lda fa			; get device number
	sta fat,x		; save to device number table
	beq lb237		; do ok exit if keyboard
	cmp #$03		; compare device number with screen
	beq lb237		; do ok exit if screen
	bcc lb234		; branch if < screen, tape or RS-232
	jmp lb22f		; skip dead code

	nop			; *** dead code ***
	jmp $f43f		; *** dead code ***
	cmp #$09		; *** dead code ***
	bne lb22f		; *** dead code ***
	jmp lb5e3		; *** dead code ***

lb22f	jsr lb3d2		; send secondary address and filename
	clc			; flag ok
	rts
lb234	jmp $f444		; jump to KERNAL to handle RS-232 and tape
lb237	clc			; flag ok
	rts


;***********************************************************************************;
;
; initialize cartridge I/O registers

intevia	lda #$00		; T1 & T2 1 shot, shift disabled, PA & PB latch disabled
	sta viabacr		; set VIA B ACR
	sta viaaacr		; set VIA A ACR
	sta viabifr		; ??? clear VIA B interrupts ???
	sta viaaifr		; ??? clear VIA A interrupts ???
	lda #$ff		; CB2 high, CB1 +ve edge, CA2 high, CA1 +ve edge
	sta viabpcr		; set VIA B PCR
	sta viaapcr		; set VIA A PCR
	lda #$80		; ??? enable interrupts, no bits set ???
	sta viabier		; modify VIA B interrupts
	sta viaaier		; modify VIA A interrupts
	lda #$07		; IIIIIOOO, ATN in, NDAC in, NRFD in, DAV in,
				; EOI in, NDAC out, NRFD out, DAV out
	sta viaaddrb		; set VIA A DDRB
	sta viaapb		; set VIA A DRB outputs high
	lda #$ff		; all outputs
	sta viabddra		; set VIA B DDRA, DIO1-8
	sta viaaddra		; set VIA A DDRA (unused)
	sta viabpa2		; set VIA B DRA, all ones
	sta viaapa2		; set VIA A DRA (unused)
	lda #$00		; all inputs
	sta viabddrb		; set VIA B DDRB, DIO1-8
	rts

	lda #$fd		; *** dead code ***
	sta bsour		; *** dead code ***
	lda bsour		; *** dead code ***
	sta unkwn2		; *** dead code ***
lb27c	lda #$40		; *** dead code ***
	sta unkwn2		; *** dead code ***
	lda bsour		; *** dead code ***
	sta unkwn2		; *** dead code ***
	lda #$12		; *** dead code ***
	sta unkwn2		; *** dead code ***
	rts			; *** dead code ***

; OR into I/O status byte

eoriost	ora status		; OR with I/O status byte
	sta status		; save I/O status byte
	rts

lb291	lda viaapb		; get VIA A DRB
	ora #$02		; set NRFD out high
	sta viaapb		; set VIA A DRB
	lda #$ff		; set for 65,280 cycles
	sta viabt1ch		; set VIA B T1C_h
lb29e	lda viaapb		; get VIA A DRB
	and #$10		; mask DAV in
	beq lb2b1		; branch if DAV low
	lda viabifr		; get VIA B IFR
	and #$40		; mask T1 interrupt
	beq lb29e		; loop while timer not expired
	lda #$02		; set time out read
	sta status		; set I/O status byte
	rts

lb2b1	lda viaapb		; get VIA A DRB
	and #$08		; mask EOI in
	bne lb2be		; branch if not EOI
	lda #$40		; set EOI bit
	jsr eoriost		; OR into I/O status byte
	nop
lb2be	lda viabpb		; get VIA B DRB
	eor #$ff		; invert it
	sta bsour		; save byte received
	lda viaapb		; get VIA A DRB
	and #$fd		; mask NRFD out
	sta viaapb		; store VIA A DRB
	lda viaapb		; get VIA A DRB
	ora #$04		; set NDAC out
	sta viaapb		; store VIA A DRB
lb2d5	lda viaapb		; get VIA A DRB
	and #$10		; mask DAV in
	beq lb2d5		; loop if DAV low
	lda viaapb		; get VIA A DRB
	and #$f9		; mask NRFD and NDAC out
	sta viaapb		; store VIA A DRB
	rts

; output byte on IEEE bus

lb2e5	lda bsour		; get byte to output
	eor #$ff		; invert it
	sta viabpa2		; set VIA B DRA
lb2ec	lda viaapb		; get VIA A DRB
	and #$20		; mask NRFD in
	beq lb2ec		; branch if NRFD low
	lda viaapb		; get VIA A DRB
	and #$fe		; clear DAV out
	sta viaapb		; store VIA A DRB
	lda #$ff		; set for 65,280 cycles
	sta viabt1ch		; set VIA B T1C_h
lb300	lda viaapb		; get VIA A DRB
	and #$40		; mask NDAC in
	bne lb313		; branch if NDAC high
	lda viabifr		; get VIA B IFR
	and #$40		; mask T1 interrupt
	beq lb300		; loop while timer not expired
	lda #$01		; set time out write
	sta status		; save I/O status byte
	rts
lb313	lda viaapb		; get VIA B IFR
	ora #$01		; set DAV out high
	sta viaapb		; store VIA A DRB
	lda #$ff		; set data to $00
	sta viabpa2		; set VIA B DRA
	clc			; flag ok
	rts

lb322	lda viabpcr		; get VIA B PCR
	and #$df		; set CB2 low, EOI low
	sta viabpcr		; save VIA B PCR
	jsr lb2e5		; output a byte to the IEEE bus
	lda viabpcr		; get VIA B PCR
	ora #$20		; set CB2 high, EOI high
	sta viabpcr		; save VIA B PCR
	rts


;***********************************************************************************;
;
; command an IEEE bus device to TALK

; To use this routine the accumulator must first be loaded with a device number
; between 4 and 30. When called this routine converts this device number to a talk
; address. Then this data is transmitted as a command on the IEEE bus.

lb336	ora #$40		; OR with the TALK command
	.byte $2c		; makes next line BIT $2009


;***********************************************************************************;
;
; command devices on the IEEE bus to LISTEN

; This routine will command a device on the IEEE bus to receive data. The
; accumulator must be loaded with a device number between 4 and 31 before calling
; this routine. LISTEN convert this to a listen address then transmit this data as
; a command on the IEEE bus. The specified device will then go into listen mode
; and be ready to accept information.

lb339	ora #$20		; OR with the LISTEN command
lb33b	sta bsour		; store byte to send
	jsr lb381		; set ATN low and wait for NDAC to got low
	jsr lb2e5		; output a byte to the IEEE bus
	rts

lb344	sta bsour		; store byte to send
	jsr lb2e5		; output a byte to the IEEE bus
	jsr lb398		; set ATN high
	rts

lb34d	tax			; copy device number to .X
	jsr lb336		; command an IEEE bus device to TALK
	lda sa			; get secondary address
	bpl lb35b		; branch if address to send
	nop
	nop
	nop
	jmp lb366		; check device is present
lb35b	sta bsour		; store byte to send
	jsr lb2e5		; output a byte to the IEEE bus
	nop
	nop
	nop
	nop
	nop
	nop
lb366	lda viaapb		; get VIA B DRB
	and #$f9		; mask NRFD and NDAC out
	sta viaapb		; save VIA B DRB
	ldy #$10		; set delay loop interations
lb370	dey			; decrement loop count
	bne lb370		; branch while non-zero
	jsr lb398		; set ATN high
	txa			; copy device back to .A
	bit status		; test I/O status byte
	bpl lb380		; if device present save device number and exit
	pla			; else dump calling address low byte
	pla			; dump calling address high byte
	jmp fe_dvntp		; do device not present error and return
lb380	rts

lb381	lda viabpcr		; get VIA B PCR
	and #$fd		; set CA2 low, ATN low
	sta viabpcr		; save VIA B PCR
	lda viaapb		; get VIA A DRB
	and #$60		; mask NRFD and NDAC in
	cmp #$60		; test both NRFD and NDAC high
	bne lb397		; branch if not both high
	lda #$80		; set device not present bit
	jsr eoriost		; OR into I/O status byte
lb397	rts

lb398	lda viabpcr		; get VIA B PCR
	ora #$02		; set CA2 high, ATN high
	sta viabpcr		; save VIA B PCR
	rts


;***********************************************************************************;
;
; command the IEEE bus to UNLISTEN

; This routine commands all devices on the IEEE bus to stop receiving data from
; the computer. Calling this routine results in an UNLISTEN command being transmitted
; on the IEEE bus. Only devices previously commanded to listen will be affected.

; This routine is normally used after the computer is finished sending data to
; external devices. Sending the UNLISTEN will command the listening devices to get
; off the IEEE bus so it can be used for other purposes.

lb3a1	lda #$3f		; set the UNLISTEN command
	jsr lb33b		; send command
	jsr lb398		; set ATN high
	rts

lb3aa	lda viaapb		; get VIA A DRB
	ora #$06		; set NDAC out and NRFD out high
	sta viaapb		; set VIA A DRB
	lda #$5f		; set the UNTALK command
	jsr lb33b		; send command
	jsr lb398		; set ATN high
	rts

; close the IEEE bus device

lb3bb	bit sa			; test the secondary address
	bmi lb3d0		; if already closed just exit
	lda fa			; get device number
	jsr lb339		; command devices on the IEEE bus to LISTEN
	lda sa			; get secondary address
	and #$ef		; mask the channel number
	ora #$e0		; OR with the CLOSE command
	jsr lb344		; send secondary address after LISTEN
	jsr lb3a1		; command the IEEE bus to UNLISTEN
lb3d0	clc			; flag ok
	rts


;***********************************************************************************;
;
; send secondary address and filename

lb3d2	lda sa			; get secondary address
	bmi lb40a		; ok exit if no address
	ldy fnlen		; get file name length
	beq lb40a		; ok exit if null
	lda fa			; get device number
	jsr lb339		; command devices on the IEEE bus to LISTEN
	lda sa			; get the secondary address
	ora #$f0		; OR with the OPEN command
	jsr lb344		; send secondary address after LISTEN
	lda status		; get I/O status byte
	bpl lb3ef		; branch if device present
	pla			; else dump calling address low byte
	pla			; dump calling address high byte
	jmp fe_dvntp		; do device not present error and return

lb3ef	lda fnlen		; get file name length
	beq lb407		; branch if null name
	ldy #$00		; clear index
lb3f5	lda (fnadr),y		; get file name byte
	sta bsour		; store byte to send
	iny			; increment index
	cpy fnlen		; compare with file name length
	beq lb404		; branch if loop done
	jsr lb2e5		; output a byte to the IEEE bus
	jmp lb3f5		; continue processing file name
lb404	jsr lb322		; output a byte with EOI to the IEEE bus
lb407	jsr lb3a1		; command the IEEE bus to UNLISTEN
lb40a	clc			; flag ok
	rts

lb40c	tax			; copy device number
	jsr lb339		; command devices on the IEEE bus to LISTEN
	lda sa			; get secondary address
	bpl lb41a		; branch if address to send
	jsr lb398		; set ATN high
	jmp lb41d		; no address to send
lb41a	jsr lb344		; send secondary address after LISTEN
lb41d	txa			; copy device number back to .A
	bit status		; test I/O status byte
	bpl lb427		; if device present save output device number and exit
	pla			; else dump calling address low byte
	pla			; dump calling address high byte
	jmp fe_dvntp		; do device not present error and return
lb427	rts

lb428	lda la			; get logical file
	bmi lb434		; branch if top bit set, BASIC sends [LF] after [CR]
	lda bsour		; get byte to output
	cmp #$0d		; compare with [CR]
	beq lb445		; output a byte with EOI and return
	bne lb44a		; branch always
lb434	lda bsour		; get byte to output
	cmp #$0a		; compare with [LF]
	beq lb445		; output a byte with EOI and return
	bne lb44a		; branch always

	bit $94			; *** dead code ***
	bmi lb445		; *** dead code ***
	sec			; *** dead code ***
	ror $94			; *** dead code ***
	bne lb44a		; *** dead code ***

lb445	jsr lb322		; output a byte with EOI to the IEEE bus
	clc			; flag ok
	rts
lb44a	jsr lb2e5		; output a byte to the IEEE bus
	clc			; flag ok
	rts
lb44f	jsr lb291		; input a byte from the IEEE bus
	clc			; flag ok
	rts


;***********************************************************************************;
;
; load RAM from a device

; This routine will load data bytes from any input device directly into the memory
; of the computer. It can also be used for a verify operation comparing data from a
; device with the data already in memory, leaving the data stored in RAM unchanged.

; The accumulator must be set to 0 for a load operation or 1 for a verify. If the
; input device was OPENed with a secondary address of 0 the header information from
; device will be ignored. In this case .X.Y must contain the starting address for the
; load. If the device was addressed with a secondary address of 1 or 2 the data will
; load into memory starting at the location specified by the header. This routine
; returns the address of the highest RAM location which was loaded.

; Before this routine can be called, the SETLFS and SETNAM routines must be called.

eload	sta verck		; save load/verify flag
	lda #$00		; clear .A
	sta status		; clear I/O status byte
	lda fa			; get device number
	bne lb461		; branch if not keyboard
lb45e	jmp fe_ildev		; do illegal device number and return
lb461	cmp #$03		; compare device number with screen
	beq lb45e		; if screen go do illegal device number and return
	bcc lb46e		; branch if less than screen
	jmp lb471		; jump to handle IEEE bus

	nop			; *** dead code ***
	jmp $f55c		; *** dead code ***

lb46e	jmp $f5ca		; jump to KERNAL to handle tape
lb471	ldy fnlen		; get file name length
	bne lb478		; branch if not null name
	jmp fe_misfn		; else do missing file name error and return
lb478	jsr lb502		; get secondary address and print "Searching..."
	lda #$60		; set secondary address to $00
	sta sa			; save secondary address
	jsr lb3d2		; send secondary address and filename
	lda fa			; get device number
	jsr lb336		; command an IEEE bus device to TALK
	lda sa			; get secondary address
	jsr lb344		; send secondary address after TALK
	lda viaapb		; get VIA A DRB
	and #$f9		; mask NRFD and NDAC out
	sta viaapb		; store VIA A DRB
	jsr lb44f		; input a byte from the IEEE bus
	lda bsour		; get received byte
	sta eal			; save program start address low byte
	lda status		; get I/O status byte
	lsr			; shift time out read ..
	lsr			; .. into carry bit
	bcs lb4f4		; if timed out go do file not found error and return
	jsr lb44f		; input a byte from the IEEE bus
	lda bsour		; get received byte
	sta eal+1		; save program start address high byte
	jsr lb507		; set LOAD address if secondary address = 0
lb4ab	lda #$fd		; mask xxxx xx0x, clear time out read bit
	and status		; mask I/O status byte
	sta status		; set I/O status byte
	jsr stop		; scan stop key, return Zb = 1 = [STOP]
	bne lb4c1		; branch if not [STOP]
	lda viaapb		; get VIA A DRB
	ora #$06		; set NDAC out and NRFD out high
	sta viaapb		; set VIA A DRB
	jmp $f6cb		; ??? close the serial bus device and flag stop ???
lb4c1	jsr lb44f		; input a byte from the IEEE bus
	lda bsour		; get received byte
	pha
	pla
	tax			; copy byte
	lda status		; get I/O status byte
	lsr			; shift time out read ..
	lsr			; .. into carry bit
	bcs lb4ab		; if timed out ???
	txa			; copy received byte back
	ldy verck		; get load/verify flag
	beq lb4e0		; branch if load
	ldy #$00		; clear index
	cmp (eal),y		; compare byte with previously loaded byte
	beq lb4e2		; branch if match
	lda #$10		; set read error bit
	jsr oriost		; OR into I/O status byte
	.byte $2c		; makes next line BIT $AE91
lb4e0	sta (eal),y		; save byte to memory
lb4e2	inc eal			; increment save pointer low byte
	bne lb4e8		; if no rollover skip the high byte increment
	inc eal+1		; else increment save pointer high byte
lb4e8	bit status		; test I/O status byte
	bvc lb4ab		; loop if not end of file
	jsr lb3aa		; send UNTALK command
	jsr lb3bb		; close IEEE bus device
	bcc lb4ff
lb4f4	lda viaapb		; get VIA A DRB
	ora #$06		; set NDAC out and NRFD out high
	sta viaapb		; set VIA A DRB
	jmp fe_ntfnd		; do file not found error and return
lb4ff	jmp $f641		; flag ok and exit


;***********************************************************************************;
;
; get secondary address and print "Searching..."

lb502	ldx sa			; get secondary address
	jmp srching		; print "Searching..." and return


;***********************************************************************************;
;
; set LOAD address if secondary address = 0

lb507	txa			; copy secondary address
	bne lb512		; load location not set in LOAD call, so
				; continue with load
	lda memuss		; get load start address low byte
	sta eal			; save program start address low byte
	lda memuss+1		; get load start address high byte
	sta eal+1		; save program start address high byte
lb512	jmp ldvrmsg		; display "LOADING" or "VERIFYING" and return


;***********************************************************************************;
;
; save RAM to device, .A = index to start address, .X.Y = end address low/high

; This routine saves a section of memory. Memory is saved from an indirect address
; on page 0 specified by A, to the address stored in .X.Y, to a logical file. The
; SETLFS and SETNAM routines must be used before calling this routine. However, a
; file name is not required to SAVE to device 1, the cassette. Any attempt to save to
; other devices without using a file name results in an error.

esave	lda fa			; get device number
	bne lb51c		; branch if not keyboard
lb519	jmp fe_ildev		; else do illegal device number and return
lb51c	cmp #$03		; compare device number with screen
	beq lb519		; if screen do illegal device number and return
	bcc lb529		; branch if < screen
	jmp lb533		; jump to handle IEEE bus

	nop			; *** dead code ***
	jmp $f692		; *** dead code ***

lb529	cmp #$02		; compare device with RS-232 device
	bne lb530		; branch if not RS-232 device
	jmp rsdvcerr		; else do illegal device number and return
lb530	jmp $f6f8		; jump to KERNAL to handle tape

lb533	lda #$61		; set secondary address to $01
	sta sa			; save secondary address
	ldy fnlen		; get file name length
	bne lb53e		; branch if filename not null
	jmp fe_misfn		; else do missing file name error and return
lb53e	jsr lb3d2		; send secondary address and filename
	jsr saving		; print saving [file name]
	lda fa			; get device number
	jsr lb339		; command devices on the IEEE bus to LISTEN
	lda sa			; get secondary address
	jsr lb344		; send secondary address after LISTEN
	ldy #$00		; clear index
	jsr rd300		; copy I/O start address to buffer address
	lda sal			; get buffer address low byte
	sta bsour		; store byte to send
	jsr lb2e5		; output a byte to the IEEE bus
	lda sal+1		; get buffer address high byte
	sta bsour		; store byte to send
	jsr lb2e5		; output a byte to the IEEE bus
lb561	jsr vprty		; check read/write pointer, return Cb = 1 if pointer >= end
	bcs lb57e		; go do UNLISTEN if at end
	lda (sal),y		; get byte from buffer
	sta bsour		; store byte to send
	jsr lb2e5		; output a byte to the IEEE bus
	jsr stop		; scan stop key
	bne lb579		; if stop not pressed go increment pointer and loop for next
	jsr lb3bb		; close IEEE bus device
	lda #$00
	sec			; flag stop
	rts
lb579	jsr wrt62		; increment read/write pointer
	bne lb561		; loop, branch always
lb57e	jsr lb3a1		; command the IEEE bus to UNLISTEN
	jsr lb3bb		; close IEEE bus device
	rts


;***********************************************************************************;
;
; get a character from the input device

; In practice this routine operates identically to the CHRIN routine for all devices
; except for the keyboard. If the keyboard is the current input device this routine
; will get one character from the keyboard buffer. It depends on the IRQ routine to
; read the keyboard and put characters into the buffer.

egetin	lda dfltn		; get input device number
	bne lb58c		; branch if not keyboard
	jmp $f1f9		; jump to KERNAL to handle keyboard
lb58c	cmp #$02		; compare device with RS-232 device
	bne lb593		; branch if not RS-232 device
	jmp $f205		; jump to KERNAL to handle RS-232
lb593	jmp lb59a		; skip dead code

	nop			; *** dead code ***
	jmp $f205		; *** dead code ***

lb59a	jmp lb154		; jump to ECHRIN


;***********************************************************************************;
;
; dead code

lb59d	tax
	lda #$33
	sta unkwn2
lb5a3	lda unkwn2
	and #$80
	beq lb5a3
	txa
	rts
lb5ac	sei
	lda bsour
	cmp #$0d
	bne lb5cc
	jsr lb63b
	lda bsour
	sta unkwn1
lb5bb	lda unkwn2
	and #$04
	beq lb5bb
	lda #$12
	sta unkwn2
	lda bsour
	cli
	clc
	rts
lb5cc	jsr lb63b
	lda bsour
	sta unkwn1
	cli
	clc
	rts
	lda #$32
	sta unkwn2
	rts
	lda #$32
	sta unkwn2
	rts
lb5e3	ldy #$00
	sty $0297
	cpy fnlen
	beq lb5f3
	lda (fnadr),y
	sta bsour
	jsr lb27c
lb5f3	jmp $f51b
lb5f6	pla
	jsr $f3b2
	lda #$fd
	sta bsour
	jsr lb27c
	sei
	lda #<nmi2
	sta nminv
	lda #>nmi2
	sta nminv+1
	cli
	jmp $f377
lb610	sta dfltn
	sei
	lda #<lb625
	sta nminv
	lda #>lb625
	sta nminv+1
	cli
	lda #$36
	sta unkwn2
	clc
	rts
lb625	pha
	txa
	pha
	tya
	pha
	lda unkwn1
	ldy $029b
	sta ($f7),y
	inc $029b
	pla
	tay
	pla
	tax
	pla
	rti
lb63b	lda unkwn2
	and #$01
	beq lb63b
	rts
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa
	lsr $94
	jmp $ff8a
	.byte $aa,$aa,$aa,$aa,$aa
	bit $94
	bpl lb7b0
	sec
	ror $a3
	jsr $ee49
	lsr $94
	lsr $a3
lb7b0	jsr $b239
	jmp $b034
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa,$aa,$aa,$aa,$aa,$aa,$aa
	.byte $aa,$aa
